"""
This provides lazy module wrappers such that old-style RETURNN-as-a-framework imports are still working.
E.g. like::

    import returnn.TFUtil
    import returnn.TFNativeOp

Some old RETURNN configs also might directly have imports like::

    import TFUtil
    import TFHorovod

This is supported as well.

"""

import sys
import os
import types
import typing
import importlib

old_to_new_mod_mapping = {
    "rnn": "__main__",
    "better_exchook": "util.better_exchook",
    "BundleFile": "datasets.bundle_file",
    "CachedDataset": "datasets.cached",
    "CachedDataset2": "datasets.cached2",
    "Config": "config",
    "Dataset": "datasets.basic",
    "DebugHelpers": "util.debug_helpers",
    "Debug": "util.debug",
    "EngineBase": "engine.base",
    "EngineBatch": "engine.batch",
    "Fsa": "util.fsa",
    "GeneratingDataset": "datasets.generating",
    "HDFDataset": "datasets.hdf",
    "HyperParamTuning": "tf.hyper_param_tuning",
    "LearningRateControl": "learning_rate_control",
    "LmDataset": "datasets.lm",
    "Log": "log",
    "MetaDataset": "datasets.meta",
    "NativeOp": "native_op",
    "NormalizationData": "datasets.normalization_data",
    "NumpyDumpDataset": "datasets.numpy_dump",
    "Pretrain": "pretrain",
    "RawWavDataset": "datasets.raw_wav",
    "SprintCache": "sprint.cache",
    "SprintControl": "sprint.control",
    "SprintDataset": "datasets.sprint",
    "SprintErrorSignals": "sprint.error_signals",
    "SprintExternInterface": "sprint.extern_interface",
    "SprintInterface": "sprint.interface",
    "StereoDataset": "datasets.stereo",
    "TaskSystem": "util.task_system",
    "TFCompat": "tf.compat",
    "TFDataPipeline": "tf.data_pipeline",
    "TFDistributed": "tf.distributed",
    "TFEngine": "tf.engine",
    "TFHorovod": "tf.horovod",
    "TFKenLM": "tf.util.ken_lm",
    "TFNativeOp": "tf.native_op",
    "TFNetwork": "tf.network",
    "TFNetworkLayer": "tf.layers.basic",
    "TFNetworkRecLayer": "tf.layers.rec",
    "TFNetworkSegModLayer": "tf.layers.segmental_model",
    "TFNetworkSigProcLayer": "tf.layers.signal_processing",
    "TFSprint": "tf.sprint",
    "TFUpdater": "tf.updater",
    "TFUtil": "tf.util.basic",
    "Util": "util.basic",
}

_base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
_mod_cache = {}  # new/old mod name -> mod


def setup(package_name=__package__, modules=None):
    """
    This does the setup, such that all the modules become available in the `returnn` package.
    It does not import all the modules now, but instead provides them lazily.

    :param str package_name: "returnn" by default
    :param dict[str,types.ModuleType]|None modules: if set, will do ``modules[old_mod_name] = mod``
    """
    for old_mod_name, new_mod_name in sorted(old_to_new_mod_mapping.items()):
        full_mod_name = "returnn.%s" % new_mod_name
        full_old_mod_name = "%s.%s" % (package_name, old_mod_name)
        if full_mod_name in _mod_cache:
            mod = _mod_cache[full_mod_name]
        elif full_mod_name in sys.modules:
            mod = sys.modules[full_mod_name]
            _mod_cache[full_mod_name] = mod
        else:
            mod = _LazyLoader(
                full_mod_name=full_mod_name,
                full_old_mod_name=full_old_mod_name,
                old_mod_name=old_mod_name,
                modules=modules,
            )
        if old_mod_name not in sys.modules:
            sys.modules[old_mod_name] = mod
        if full_old_mod_name not in sys.modules:
            sys.modules[full_old_mod_name] = mod
        if modules is not None:
            modules[old_mod_name] = mod


class _LazyLoader(types.ModuleType):
    """
    Lazily import a module, mainly to avoid pulling in large dependencies.
    Code borrowed from TensorFlow, and simplified, and extended.
    """

    def __init__(self, full_mod_name, **kwargs):
        """
        :param str full_mod_name:
        """
        super(_LazyLoader, self).__init__(full_mod_name)
        fn = "%s/%s.py" % (_base_dir, full_mod_name.replace(".", "/"))
        if not os.path.exists(fn):
            fn = "%s/%s/__init__.py" % (_base_dir, full_mod_name.replace(".", "/"))
            assert os.path.exists(fn), "_LazyLoader: mod %r not found in %r" % (full_mod_name, _base_dir)
        self.__file__ = fn
        self._lazy_mod_config = dict(full_mod_name=full_mod_name, **kwargs)  # type: typing.Dict[str]

    def _load(self):
        full_mod_name = self.__name__
        lazy_mod_config = self._lazy_mod_config
        old_mod_name = lazy_mod_config.get("old_mod_name", None)
        full_old_mod_name = lazy_mod_config.get("full_old_mod_name", None)
        modules = lazy_mod_config.get("modules", None)
        if full_mod_name in _mod_cache:
            module = _mod_cache[full_mod_name]
        else:
            try:
                module = importlib.import_module(full_mod_name)
            except Exception:
                # Note: If we get any exception in the module itself (e.g. No module named 'theano' or so),
                # just pass it on. But this can happen.
                raise
            _mod_cache[full_mod_name] = module

        if old_mod_name:
            sys.modules[old_mod_name] = module
        if full_old_mod_name:
            sys.modules[full_old_mod_name] = module
        if modules is not None:
            assert old_mod_name
            modules[old_mod_name] = module

        # Do not set self.__dict__, because the module itself could later update itself.
        return module

    def __getattribute__(self, item):
        # Implement also __getattribute__ such that early access to just self.__dict__ (e.g. via vars(self)) also works.
        if item == "__dict__":
            # noinspection PyBroadException
            try:
                mod = self._load()
            except Exception:  # many things could happen
                print("WARNING: %s cannot be imported, __dict__ not available" % self.__name__)
                # In many cases, this is not so critical, because we likely just checked the dict content or so.
                # This should be safe, as we have this registered in sys.modules, and some code just iterates
                # through all sys.modules to check for something.
                # Any other attribute access will lead to the real exception.
                # We ignore this for __dict__, and just return a dummy empty dict.
                return {}
            return getattr(mod, "__dict__")
        return super(_LazyLoader, self).__getattribute__(item)

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)

    def __setattr__(self, key, value):
        if key in ["__file__", "_lazy_mod_config"]:
            super(_LazyLoader, self).__setattr__(key, value)
            return
        module = self._load()
        setattr(module, key, value)



"""
Base class for Tensor mixins. Just to define the attribs.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Union, Tuple

if TYPE_CHECKING:
    from ._tensor_extra import _TensorExtra
    from returnn.util.basic import NotSpecified
    import returnn.tensor.tensor as _t
    from .dim import Dim


class _TensorMixinBase:
    name: str
    _dims: Tuple[Dim, ...]
    dtype: str
    sparse_dim: Optional[Dim]
    _feature_dim_axis: Optional[Union[int, NotSpecified]]
    _raw_tensor: Optional[_t.RawTensorType]
    raw_tensor: Optional[_t.RawTensorType]
    version: int
    _extra: Optional[_TensorExtra]



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_pack_padded():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            pack, pack_dim = rf.pack_padded(x, dims=[batch_dim, time_dim], enforce_sorted=False)
            return pack, pack_dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, pack_dim = model(extern_data["data"])
        out.mark_as_default_output(shape=(pack_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_reshape():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.reshape(x, in_dims=(time_dim, in_dim), out_dims=(in_dim, time_dim))

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_expand_dim():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    expand_dim = Dim(3, name="expand")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.expand_dim(x, expand_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, expand_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_concat():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            return rf.concat((x, in_dim), (x, in_dim))

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, dim = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pad():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim, Dim]]:
            pack, (new_time, new_feat) = rf.pad(x, axes=[time_dim, in_dim], padding=[(1, 2), (3, 4)], value=0)
            return pack, (new_time, new_feat)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, (new_time, new_feat) = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, new_feat))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pad_time():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim, Dim]]:
            pack, (new_time,) = rf.pad(x, axes=[time_dim], padding=[(1, 0)], value=0)
            return pack, (new_time,)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, (new_time,) = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_gather():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.gather(x, indices=0, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_gather_2d_indices():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=in_dim),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor, y: Tensor) -> Tensor:
            return rf.gather(x, indices=y, axis=in_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], extern_data["classes"])
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_slice():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            pack, new_time = rf.slice(x, axis=time_dim, start=1, size=2)
            return pack, new_time

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, new_time = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_shift_right():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], sparse_dim=in_dim, dtype="int32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.shift_right(x, axis=time_dim, pad_value=0)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_reverse_sequence():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.reverse_sequence(x, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_where():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "cond": Tensor("cond", [batch_dim, time_dim], dtype="bool"),
            "true": Tensor("true", [batch_dim, time_dim, in_dim], dtype="float32"),
            "false": Tensor("false", [batch_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = rf.where(extern_data["cond"], extern_data["true"], extern_data["false"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)



"""
Loss functions
"""

from __future__ import annotations
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


def cross_entropy(
    *,
    estimated: Tensor,
    target: Tensor,
    axis: Dim,
    estimated_type: str,
) -> Tensor:
    """
    ``target`` is supposed to be in probability space (normalized). It can also be sparse, i.e. contain class indices.
    ``estimated`` can be probs, log-probs or logits, specified via ``estimated_type``.

    Assuming both are in probability space, the cross entropy is:

        H(target,estimated) = -reduce_sum(target * log(estimated), axis=axis)
                            = -matmul(target, log(estimated), reduce=axis)

    In case you want label smoothing, you can use e.g.::

        ce = nn.cross_entropy(
            target=nn.label_smoothing(target, 0.1),
            estimated=estimated)

    :param estimated: probs, log-probs or logits, specified via ``estimated_type``
    :param target: probs, normalized, can also be sparse
    :param axis: class labels dim over which softmax is computed
    :param estimated_type: "probs", "log-probs" or "logits"
    :return: cross entropy (same Dims as 'estimated' but without 'axis')
    """

    if estimated_type == "logits":
        # This is a common case and most backends provide optimized functions for it.
        # noinspection PyProtectedMember
        return estimated._raw_backend.softmax_cross_entropy_with_logits(logits=estimated, targets=target, axis=axis)
    if estimated_type == "probs":
        log_prob = rf.log(estimated)  # TODO: make numerically stable
    elif estimated_type == "log-probs":
        log_prob = estimated
    else:
        raise ValueError("estimated_type must be 'probs', 'log-probs' or 'logits'")
    if target.sparse_dim:
        return -rf.gather(log_prob, indices=target, axis=axis)
    return -rf.matmul(target, log_prob, reduce=axis)



"""
Tests for PyTorch engine.
"""

import _setup_test_env  # noqa
import torch

from returnn.config import Config, global_config_ctx
from returnn.tensor import TensorDict, Tensor
from returnn.torch.engine import Engine
import returnn.frontend as rf
from returnn.forward_iface import ForwardCallbackIface
from returnn.datasets import init_dataset


def test_torch_engine():
    def _get_model(**_kwargs):
        return torch.nn.Module()

    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        rf.get_run_ctx().mark_as_default_output(extern_data["data"])

    class _ForwardCallback(ForwardCallbackIface):
        def __init__(self):
            self.num_seqs = 0
            self.init_called = False
            self.finish_called = False

        def init(self, *, model):
            assert isinstance(model, torch.nn.Module)
            assert self.num_seqs == 0
            self.init_called = True

        def process_seq(self, *, seq_tag: str, outputs: TensorDict):
            out = outputs["output"]
            assert isinstance(out, Tensor)
            assert out.batch_ndim == 2 and out.batch_shape[-1] == 9
            self.num_seqs += 1

        def finish(self):
            self.finish_called = True

    config = Config(
        dict(task="forward", extern_data={"data": {"dim": 9}}, get_model=_get_model, forward_step=_forward_step)
    )
    dataset = init_dataset({"class": "Task12AXDataset", "num_seqs": 100, "name": "dev", "fixed_random_seed": 1})
    callback = _ForwardCallback()

    with global_config_ctx(config):
        engine = Engine(config=config)
        engine.init_network_from_config()
        engine.forward_with_callback(callback=callback, dataset=dataset)
        assert callback.num_seqs == 100
        assert callback.init_called and callback.finish_called


def test_min_seq_len():

    from returnn.datasets.generating import DummyDataset

    config = Config({"min_seq_length": 2})
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=1)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        assert False, "Should not contain sequences"

    config = Config()
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=3)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        return
    assert False, "Should have contained sequences"


def test_max_seq_len():

    from returnn.datasets.generating import DummyDataset

    config = Config({"max_seq_length": 4})
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=5)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        assert False, "Should not contain sequences"

    config = Config()
    dataset = DummyDataset(input_dim=1, output_dim=4, num_seqs=1, seq_len=3)
    engine = Engine(config=config)
    data_loader = engine._create_data_loader(dataset)
    for _ in data_loader:
        return
    assert False, "Should have contained sequences"
